#!/usr/bin/env python
#-*-coding: utf-8 -*-

import errno, os, sys, stat, signal, datetime, hashlib, json, paramiko, re, requests, ssl, socket, time, threading, traceback, urllib2, urlparse, uuid

from httplib import BadStatusLine
from pyVmomi import vim, vmodl
from pyVim.task import WaitForTask, TaskBlocked
from pyVim.connect import SmartConnect, Disconnect

from requests.packages.urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

def isSuccessStatusCode(code):
    return 200 <= code and code < 300

# copied from include/vmbk/defines.h

VSPHERE_SERVER_SOCKET_PATH = '/tmp/synoabk/vm/vsphere_server.sock'
SNAPSHOT_NAME_PREFIX = 'CREATED_FOR_SYNOLOGY_ACTIVE_BACKUP_FOR_BUSINESS'
SNAPSHOT_DESC = 'This snapshot was created by Synology Active Backup for Business. Please do not remove it if you are not sure what you are doing.'

# load vix error text

vixErrorText = {}
dsmPackagePath = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
errorsFilePath = os.path.join(dsmPackagePath, 'lib', 'vddk65', 'errors.html')

with open(errorsFilePath) as file:
    content = file.read()

matches = re.findall(r'<dt.*?>(\d+).*?<dd>(.*?)</dd>', content, re.DOTALL)

for match in matches:
    vixErrorText[match[0]] = match[1]

class GuestNotEnoughSpace(Exception):
    def __init__(self, msg):
        self.msg = msg
    def __str__(self):
        return self.msg

# vsphere server class
class vSphereServer:
    def __init__(self, hostname, username, password, thumbprint, port):
        self.si = None
        self.task_pool = {}
        self.api_type = None
        self.api_version = None
        self.auth_info = {
            'hostname'  : hostname,
            'username'  : username,
            'password'  : password,
            'thumbprint': thumbprint,
            'port'      : port
        }

        self.lock = threading.Lock()
        self.lockNum = 0
        self.lockNumLock = threading.Lock()
        self.error = None

    def __del__(self):
        Disconnect(self.si)
        print('disconnect %s' % self.auth_info['hostname'])

    def dispatch(self, command, argument):
        command = getattr(self, command)
        argument = argument or {}
        response = {}

        try:
            self.connect()
            response['data'] = command(**argument)
            response['success'] = True
        except Exception as e:
            traceback.print_exc(file=sys.stderr)
            response = self._parse_exception(e)
            response['success'] = False

        return response

    def _parse_exception(self, e):
        error_class = type(e).__name__
        error_message = ''

        if type(e) == socket.error:
            if e[0] in (errno.EHOSTUNREACH, errno.ECONNREFUSED, errno.ETIMEDOUT, errno.EHOSTDOWN, errno.EPIPE):
                error_class = 'vim.fault.HostConnectFault'
            error_message = e[1]
        elif type(e) in (BadStatusLine, ssl.SSLError, socket.timeout, requests.ConnectionError):
            error_class = 'vim.fault.HostConnectFault'
            error_message = 'Unknown connection error'
        elif type(e) == vmodl.fault.SystemError:
            m = re.search(r'vix error codes = \((.*?),', e.msg)
            if m != None and m.lastindex == 1:
                error_message = vixErrorText[m.group(1)]
            else:
                error_message = e.msg
        elif hasattr(e, 'faultMessage') and len(e.faultMessage) > 0:
            error_message = e.faultMessage[0].message
        elif hasattr(e, 'msg'):
            error_message = e.msg
        else:
            error_message = str(e)

        return {
            'error_class': error_class,
            'error_message': error_message
        }

    def _connect(self):
        self.si = SmartConnect(
            host=self.auth_info['hostname'],
            user=self.auth_info['username'],
            pwd=self.auth_info['password'],
            port=self.auth_info['port'],
            thumbprint=self.auth_info['thumbprint'],
            sslContext=ssl._create_unverified_context())

        about = self.si.content.about
        self.task_pool = {}
        self.api_type = about.apiType
        self.api_version = about.apiVersion
        print('connect %s' % self.auth_info['hostname'])

    def connect(self):
        try:
            self.lockNumLock.acquire()
            self.lockNum += 1
            self.lockNumLock.release()

            self.lock.acquire()

            if self.error:
                raise self.error
            if self.si == None:
                self._connect()
            try:
                self.si.CurrentTime()
            except vim.fault.NotAuthenticated:
                print('reauth %s' % self.auth_info['hostname'])
                self._connect()
        except Exception as e:
            self.error = e
            raise
        finally:
            self.lockNumLock.acquire()
            self.lockNum -= 1
            if self.lockNum == 0:
                self.error = None
            self.lockNumLock.release()

            self.lock.release()

    def getServerInfo(self, reconnect=False):
        def _is_shell_running(hosts):
            for host in hosts:
                for service in host.configManager.serviceSystem.serviceInfo.service:
                    if (service.key == 'TSM' or service.key == 'TSM-SSH') and service.running == False:
                        return False
            return True

        if reconnect:
            si = SmartConnect(
                host=self.auth_info['hostname'],
                user=self.auth_info['username'],
                pwd=self.auth_info['password'],
                port=self.auth_info['port'],
                thumbprint=self.auth_info['thumbprint'],
                sslContext=ssl._create_unverified_context())
            Disconnect(si)

        data = {}
        data['api_type'] = self.api_type
        data['api_version'] = self.api_version
        data['licensed'] = (self.api_type == 'VirtualCenter') or (self.si.content.licenseManager.licenses[0].total > 0)
        data['shell_available'] = data['licensed'] or _is_shell_running(self._get_objs([vim.HostSystem]))
        data['guest_available'] = data['licensed'] and (int(self.api_version.split('.')[0]) >= 5)
        return data

    def getServerCert(self):
        context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock = context.wrap_socket(sock, server_hostname=self.auth_info['hostname'])
        sock.connect((self.auth_info['hostname'], self.auth_info['port']))
        cert = sock.getpeercert(True)
        sock.close()

        sha1 = hashlib.sha1()
        sha1.update(cert)
        digest = sha1.hexdigest().lower()

        s, i = '', 0
        for c in digest:
            s += c
            i += 1
            if i % 2 == 0:
                s += ':'

        return {'thumbprint': s[:-1]}

    def getHostInfo(self, host_id):
        host = self._get_obj_by_moid(host_id, [vim.HostSystem])
        return self._parse_host_data(host)

    def getVmInfo(self, vm_id, detail=True, get_disk=True):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        return self._parse_vm_data(vm, detail, get_disk)

    def getDatastoreInfo(self, datastore_infos):
        ret = []
        for datastore_info in datastore_infos:
            datastore = self._get_obj_by_moid(datastore_info['datastore_id'], [vim.Datastore])
            ret.append(self._parse_datastore_data(datastore))
        return ret

    def getEntityPath(self, entity=None, entityId=None, viewType=None, attr='_moId'):
        d = {'entity': entity or self._get_obj_by_moid(entityId), 'path': ''}

        def _get_datacenter_path():
            while d['entity'] != self._get_root_obj():
                d['path'] = '/' + getattr(d['entity'], attr) + d['path']
                d['entity'] = d['entity'].parent

        def _get_resource_pool_path():
            while type(d['entity'].parent) != vim.ComputeResource and \
                  type(d['entity'].parent) != vim.ClusterComputeResource:
                d['path'] = '/' + getattr(d['entity'], attr) + d['path']

                if viewType == 'vim.VirtualMachine-Folder' and \
                    type(d['entity']) == vim.VirtualApp and \
                    d['entity'].parentFolder != None:
                    d['entity'] = d['entity'].parentFolder
                    return
                else:
                    d['entity'] = d['entity'].parent
            d['entity'] = d['entity'].parent

        def _get_folder_path():
            while type(d['entity'].parent) != vim.Datacenter:
                d['path'] = '/' + getattr(d['entity'], attr) + d['path']
                d['entity'] = d['entity'].parent
            d['entity'] = d['entity'].parent

        if type(d['entity']) == vim.HostSystem:
            if type(d['entity'].parent) == vim.ClusterComputeResource:
                d['path'] = '/' + getattr(d['entity'], attr)
            d['entity'] = d['entity'].parent
            _get_folder_path()

        elif type(d['entity']) == vim.ResourcePool or type(d['entity']) == vim.VirtualApp:
            _get_resource_pool_path()
            _get_folder_path()

        elif type(d['entity']) == vim.Folder:
            _get_folder_path()

        elif type(d['entity']) == vim.VirtualMachine:
            d['path'] = '/' + getattr(d['entity'], attr)

            if d['entity'].summary.config.template:
                return d['path']

            elif viewType == 'vim.VirtualMachine-Folder':
                if d['entity'].parent == None:
                    d['entity'] = d['entity'].resourcePool
                    _get_resource_pool_path()
                else:
                    d['entity'] = d['entity'].parent
                _get_folder_path()

            elif viewType == 'vim.VirtualMachine-Host':
                d['entity'] = d['entity'].resourcePool
                _get_resource_pool_path()
                _get_folder_path()

        _get_datacenter_path()
        return d['path']

    def listHypervisor(self, targets=[]):
        vms = self._get_objs([vim.VirtualMachine], self._get_root_obj())
        out = []
        for vm in vms:
            if not targets or vm._moId in targets:
                out.append({
                    'vm_id': vm._moId,
                    'host_id': vm.summary.runtime.host._moId
                })
        return out

    def listChildEntity(self, parentId='', viewType='', recursive=False, getVmDisk=False, getVmView=False):
        def _parse_data(child):
            if type(child) == vim.VirtualMachine:
                return self._parse_vm_data(child, False, getVmDisk, getVmView, viewCache)
            elif type(child) == vim.Datastore:
                return self._parse_datastore_data(child)
            elif type(child) == vim.Datacenter:
                return {
                    'name': child.name,
                    'folder_id': child.vmFolder._moId
                }
            elif type(child) == vim.ComputeResource:
                return {
                    'name': child.name,
                    'host_id': child.host[0]._moId,
                    'resourcepool_id': child.resourcePool._moId,
                    'migratable': self.getMigrateCapability(host_obj=child.host[0])
                }
            elif type(child) == vim.ClusterComputeResource:
                return {
                    'name': child.name,
                    'resourcepool_id': child.resourcePool._moId
                }
            elif type(child) == vim.HostSystem:
                return {
                    'name': child.name,
                    'migratable': self.getMigrateCapability(host_obj=child)
                }
            elif type(child) == vim.VirtualApp:
                return {
                    'name': child.name,
                    'started': child.summary.vAppState == 'started'
                }
            else:
                return {'name': child.name}

        children = []
        entities = []
        viewCache = {}

        if parentId == '':
            parentObj = self._get_root_obj()
        else:
            parentObj = self._get_obj_by_moid(parentId)

        if parentId == '':
            if viewType == 'vim.VirtualMachine':
                children = self._get_objs([vim.VirtualMachine], parentObj)
            elif viewType == 'vim.HostSystem':
                children = self._get_objs([vim.HostSystem], parentObj)
            else:
                children = parentObj.childEntity

        elif type(parentObj) == vim.Folder:
            if viewType == 'vim.Folder' and 'VirtualMachine' in parentObj.childType:
                children = [obj for obj in parentObj.childEntity if type(obj) == vim.Folder]
            elif viewType == 'vim.VirtualMachine':
                children = [obj for obj in parentObj.childEntity if type(obj) == vim.VirtualMachine]
            else:
                children = parentObj.childEntity

        elif type(parentObj) == vim.Datacenter:
            if viewType == 'vim.Folder':
                if recursive:
                    children = [parentObj.vmFolder] + self._get_objs([vim.Folder], parentObj.vmFolder)
                else:
                    children = [obj for obj in parentObj.vmFolder.childEntity if type(obj) == vim.Folder]
            elif viewType == 'vim.VirtualMachine-Folder':
                children = parentObj.vmFolder.childEntity
            elif viewType == 'vim.ResourcePool':
                if recursive:
                    children = self._get_objs([vim.ResourcePool], parentObj.hostFolder)
                else:
                    children = parentObj.hostFolder.childEntity
            else:
                if recursive:
                    children = self._get_objs([vim.HostSystem], parentObj.hostFolder)
                else:
                    children = parentObj.hostFolder.childEntity

        elif type(parentObj) == vim.HostSystem:
            if viewType == 'vim.Datastore':
                children = parentObj.datastore
            elif viewType == 'vim.Network':
                children = parentObj.network
            else:
                children = []

        elif type(parentObj) == vim.ClusterComputeResource:
            children = parentObj.host

            if viewType == 'vim.VirtualMachine-Host':
                children += parentObj.resourcePool.vm

        elif type(parentObj) == vim.ComputeResource:
            if viewType == 'vim.Datastore':
                children = parentObj.datastore
            elif viewType == 'vim.Network':
                children = parentObj.network
            else:
                children = parentObj.resourcePool.resourcePool

                if viewType == 'vim.VirtualMachine-Host':
                    children += parentObj.resourcePool.vm

        elif type(parentObj) == vim.ResourcePool or type(parentObj) == vim.VirtualApp:
            children = parentObj.resourcePool

            if viewType == 'vim.VirtualMachine-Folder' or viewType == 'vim.VirtualMachine-Host':
                children += parentObj.vm

        for child in children:
            data = _parse_data(child)
            data['id'] = child._moId
            data['type'] = type(child).__name__
            entities.append(data)

        entities.sort(key=lambda obj: (obj['type'], obj['name']))
        return entities

    def renameVm(self, vm_id=None, vm_name=None):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        task = vm.Rename_Task(newName=vm_name)
        self._wait_for_task(task)

    def removeVm(self, vm_id=None):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        task = vm.Destroy_Task()
        self._wait_for_task(task)

    def registerVm(self, vm_name=None, folder_id=None, file_path=None, resourcepool_id=None, host_id=None):
        rp_obj = self._get_obj_by_moid(resourcepool_id, [vim.ResourcePool])
        host_obj = self._get_obj_by_moid(host_id, [vim.HostSystem])

        def _is_vapp(obj):
            while type(obj) == vim.ResourcePool:
                obj = obj.parent
            return type(obj) == vim.VirtualApp

        if _is_vapp(rp_obj):
            task = rp_obj.RegisterChildVM_Task(
                    path=file_path,
                    name=vm_name,
                    host=host_obj)
        else:
            folder_obj = self._get_obj_by_moid(folder_id, [vim.Folder])
            task = folder_obj.RegisterVM_Task(
                    path=file_path,
                    name=vm_name,
                    pool=rp_obj,
                    host=host_obj,
                    asTemplate=False)

        vm = self._wait_for_task(task)
        return vm._moId

    def uploadFile(self, datacenter_id=None, datastore_id=None, src_path=None, dst_path=None):
        dc_obj = self._get_obj_by_moid(datacenter_id, [vim.Datacenter])
        ds_obj = self._get_obj_by_moid(datastore_id, [vim.Datastore])
        dc_path = self._get_datacenter_path(dc_obj)

        url = 'https://%s:%s/folder/%s?dcPath=%s&dsName=%s' % (self.auth_info['hostname'], self.auth_info['port'], dst_path, dc_path, ds_obj.name)
        with open(src_path.encode('utf-8'), 'rb') as f:
            resp = requests.put(self._url_fix(url), data=f, auth=(self.auth_info['username'], self.auth_info['password']), verify=False)
            if isSuccessStatusCode(resp.status_code) == False:
                raise vim.fault.HostConnectFault(msg='Error while uploading file, status code: %d' % resp.status_code)

    def createNfsDatastore(self, name=None, host_id=None, ip_address=None, mount_path=None):
        host_obj = self._get_obj_by_moid(host_id, [vim.HostSystem])
        host_datastore_system = host_obj.configManager.datastoreSystem
        volume_spec = vim.host.NasVolume.Specification()
        volume_spec.remoteHost = ip_address
        volume_spec.remotePath = mount_path
        volume_spec.localPath = name
        volume_spec.accessMode = 'readWrite'
        ds_obj = host_datastore_system.CreateNasDatastore(spec=volume_spec)
        return self._parse_datastore_data(ds_obj)

    def removeDatastore(self, host_id=None, datastore_id=None):
        host_obj = self._get_obj_by_moid(host_id, [vim.HostSystem])
        datastore_obj = self._get_obj_by_moid(datastore_id, [vim.Datastore])
        host_datastore_system = host_obj.configManager.datastoreSystem
        host_datastore_system.RemoveDatastore(datastore=datastore_obj)

    def createFolder(self, datacenter_id=None, folder_path=None):
        datacenter = self._get_obj_by_moid(datacenter_id, [vim.Datacenter])
        fileManager = self.si.content.fileManager
        fileManager.MakeDirectory(name=folder_path, datacenter=datacenter)

    def removeFolder(self, datacenter_id=None, folder_path=None):
        datacenter = self._get_obj_by_moid(datacenter_id, [vim.Datacenter])
        fileManager = self.si.content.fileManager
        task = fileManager.DeleteDatastoreFile_Task(name=folder_path, datacenter=datacenter)
        self._wait_for_task(task)

    def renameFolder(self, datacenter_id=None, src_path=None, dst_path=None):
        datacenter = self._get_obj_by_moid(datacenter_id, [vim.Datacenter])
        fileManager = self.si.content.fileManager
        task = fileManager.MoveDatastoreFile_Task(sourceName=src_path, destinationName=dst_path, sourceDatacenter=datacenter)
        self._wait_for_task(task)

    def createDisk(self, datacenter_id=None, file_path=None, disk_type=None, adapter_type=None, capacity=None):
        spec = vim.VirtualDiskManager.FileBackedVirtualDiskSpec()
        spec.diskType = disk_type
        spec.adapterType = adapter_type
        spec.capacityKb = int(capacity) / 1024
        dc_obj = self._get_obj_by_moid(datacenter_id, [vim.Datacenter])
        disk_manager = self.si.content.virtualDiskManager
        task = disk_manager.CreateVirtualDisk_Task(
                name=file_path,
                datacenter=dc_obj,
                spec=spec)
        return self._put_task(task)

    def getVmQuestion(self, vm_id=None):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        if vm.runtime.question != None:
            q = self._parse_question(vm.runtime.question)
            q['exist'] = True
            return q
        return {'exist': False}

    def migrateVm(self, vm_id=None, datastore_id=None):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        datastore_obj = self._get_obj_by_moid(datastore_id, [vim.Datastore])
        relocate_spec = vim.vm.RelocateSpec()
        relocate_spec.datastore = datastore_obj
        task = vm.RelocateVM_Task(spec=relocate_spec)
        return self._put_task(task)

    def poweronVm(self, vm_id=None):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        poweron_task = vm.PowerOn()
        time.sleep(10)

        if vm.runtime.question != None:
            question = self._parse_question(vm.runtime.question)
            vm.AnswerVM(question['question_id'], question['default_answer'])
            time.sleep(5)

        self._wait_for_task(poweron_task)

    def poweroffVm(self, vm_id=None):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        poweroff_task = vm.PowerOff()
        self._wait_for_task(poweroff_task)

    def unregisterVm(self, vm_id=None):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        vm.UnregisterVM()

    def listSnapshot(self, vm_id=None):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        if hasattr(vm, 'snapshot') and hasattr(vm.snapshot, 'rootSnapshotList'):
            return self._list_snapshot_recursively(vm.snapshot.rootSnapshotList)
        else:
            return []

    def getSnapshotInfo(self, vm_id=None, snap_id=None):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        if hasattr(vm, 'snapshot') and hasattr(vm.snapshot, 'rootSnapshotList'):
            snap_obj = self._get_snapshot_obj(vm.snapshot.rootSnapshotList, snap_id)
            return self._parse_snapshot_data(snap_obj)
        raise vim.fault.NotFound(msg='Cannot find [%s]' % snap_id)

    def takeSnapshot(self, vm_id=None, snap_name=None, snap_desc=None, quiesce=False):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        create_snap_task = vm.CreateSnapshot_Task(
            name=snap_name,
            description=snap_desc,
            memory=False,
            quiesce=quiesce)
        snap_obj = self._wait_for_task(create_snap_task)
        return self.getSnapshotInfo(vm_id, snap_obj._moId)

    def removeSnapshot(self, vm_id=None, snap_id=None):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        snap_obj = self._get_snapshot_obj(vm.snapshot.rootSnapshotList, snap_id)
        task = snap_obj.snapshot.RemoveSnapshot_Task(
                removeChildren=False,
                consolidate=True)
        self._wait_for_task(task)

    def getTaskInfo(self, task_id=None):
        task = self._get_task(task_id)

        if task.info.state == 'error':
            raise task.info.error
        return {
            'state': task.info.state,
            'progress': task.info.progress
        }

    def cancelTask(self, task_id=None):
        task = self._get_task(task_id)
        task.CancelTask()
        self._del_task(task_id)

    def getCbtState(self, vm_id=None):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        return {
            'support': vm.capability.changeTrackingSupported,
            'enabled': vm.config.changeTrackingEnabled
        }

    def setCbtState(self, vm_id=None, enable=False):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])

        if enable == vm.config.changeTrackingEnabled:
            return None

        cspec = vim.vm.ConfigSpec()
        cspec.changeTrackingEnabled = enable
        reconfig_vm_task = vm.ReconfigVM_Task(cspec)
        self._wait_for_task(reconfig_vm_task)

        snap_name = SNAPSHOT_NAME_PREFIX + str(datetime.datetime.now())
        snap = self.takeSnapshot(vm_id, snap_name, SNAPSHOT_DESC, False)
        self.removeSnapshot(vm_id, snap['snap_id'])

    def getDiskCbt(self, vm_id=None, snap_id=None, key=None, offset=None, change_id=None):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        snap_obj = self._get_snapshot_obj(vm.snapshot.rootSnapshotList, snap_id)
        cbt_obj = vm.QueryChangedDiskAreas(
            snapshot=snap_obj.snapshot,
            deviceKey=key,
            startOffset=offset,
            changeId=change_id)

        out_json = {}
        out_json['offset'] = cbt_obj.startOffset
        out_json['length'] = cbt_obj.length
        out_json['change_area'] = []

        for extent in cbt_obj.changedArea:
            out_json['change_area'].append({'start': extent.start, 'length': extent.length})

        for device in snap_obj.snapshot.config.hardware.device:
            if device.key == key:
                out_json['filename'] = device.backing.fileName
                break

        return out_json

    def uploadFileInVm(self, vm_id=None, vm_user=None, vm_pwd=None, upload_file=None, dst_filepath=None, overwrite=True):
        def uploadFileInVmTask(task):
            class LengthWrapper:
                def __init__(self, readable, length, task):
                    self._length = self._remaining = length
                    self._readable = readable
                    self._task = task

                def read(self, *args, **kwargs):
                    if self._task._canceled:
                        raise Exception('cancel')
                    if self._remaining <= 0:
                        return ''
                    chunk = self._readable.read(*args, **kwargs)[:self._remaining]
                    self._remaining -= len(chunk)
                    self._task.info.progress = 100.0 * (self._length - self._remaining) / self._length
                    return chunk

                def __len__(self):
                    return self._length

            try:
                vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
                creds = vim.vm.guest.NamePasswordAuthentication(username=vm_user, password=vm_pwd)

                file_size = os.path.getsize(upload_file.encode('utf-8'))
                file_attribute = vim.vm.guest.FileManager.FileAttributes()
                temp_dst_filepath = dst_filepath + '_abb.' + str(time.time())

                url = self.si.content.guestOperationsManager.fileManager. \
                        InitiateFileTransferToGuest(vm, creds, temp_dst_filepath, file_attribute, file_size, overwrite). \
                        replace('*', self.auth_info['hostname'], 1)

                with open(upload_file.encode('utf-8'), 'rb') as file:
                    lw = LengthWrapper(file, file_size, task)
                    resp = requests.put(url, data=lw, verify=False)
                if isSuccessStatusCode(resp.status_code):
                    task.info.state = 'success'
                    self.si.content.guestOperationsManager.fileManager.MoveFileInGuest(vm, creds, temp_dst_filepath, dst_filepath, overwrite)
                else:
                    raise vim.fault.HostConnectFault(msg='Error while uploading file, status code: %d' % resp.status_code)
            except Exception as e:
                if type(e) == vim.fault.FileAlreadyExists:
                    task.info.state = 'success'
                else:
                    task.info.state = 'error'
                    task.info.error = e
                    if type(e) == requests.ConnectionError and hasattr(e.args[0][1], 'errno') and e.args[0][1].errno == errno.EPIPE:
                        task.info.error = GuestNotEnoughSpace('No enough space to restore file.')
                try:
                    self.si.content.guestOperationsManager.fileManager.DeleteFileInGuest(vm, creds, temp_dst_filepath)
                except:
                    pass

        class FakeTaskInfo:
            def __init__(self):
                self.key = 'upload_to_vm_' + uuid.uuid4().hex[:12]
                self.state = 'running'
                self.error = None
                self.result = None
                self.progress = 0

        class FakeTask:
            def __init__(self):
                self.info = FakeTaskInfo()
                self._canceled = False
            def CancelTask(self):
                self._canceled = True

        task = FakeTask()
        threading.Thread(target=uploadFileInVmTask, args=(task,)).start()
        return self._put_task(task)

    def listFilesInVm(self, vm_id=None, vm_user=None, vm_pwd=None, file_path=None, index=None, max_results=None, match_pattern=None):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        auth = vim.vm.guest.NamePasswordAuthentication(username=vm_user, password=vm_pwd)
        authMgr = self.si.content.guestOperationsManager.authManager
        fileMgr = self.si.content.guestOperationsManager.fileManager
        out = {'files': [], 'remaining': 0}

        if file_path == '':
            authMgr.ValidateCredentialsInGuest(vm, auth)
            for disk in vm.guest.disk:
                out['files'].append({
                    'name': disk.diskPath.replace('\\', ''),
                    'path': disk.diskPath.replace('\\', ''),
                    'size': 0,
                    'is_dir': True,
                    'mtime': 0
                })
        else:
            resp = fileMgr.ListFilesInGuest(vm, auth, file_path, index, max_results, match_pattern)
            out['remaining'] = resp.remaining
            for file in resp.files:
                if file.path == '.' or file.path == '..':
                    continue
                out['files'].append({
                    'name': file.path,
                    'path': file_path + '/' + file.path,
                    'size': file.size,
                    'is_dir': file.type == 'directory',
                    'mtime': int(file.attributes.modificationTime.strftime('%s'))
                })

        return out

    def downloadFileFromVm(self, vm_id=None, vm_user=None, vm_pwd=None, download_file=None, dst_filepath=None):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        creds = vim.vm.guest.NamePasswordAuthentication(username=vm_user, password=vm_pwd)
        fileMgr = self.si.content.guestOperationsManager.fileManager
        fileInfo = fileMgr.InitiateFileTransferFromGuest(vm, creds, download_file)
        url = fileInfo.url.replace('*', self.auth_info['hostname'], 1)
        resp = requests.get(url, stream=True, verify=False)

        if isSuccessStatusCode(resp.status_code) == False:
            raise vim.fault.HostConnectFault(msg='Error while downloading file, status code: %d' % resp.status_code)
        with open(dst_filepath.encode('utf-8'), 'wb') as myfile:
            myfile.write(resp.text.encode('utf-8'))

    def removeFileInVm(self, vm_id=None, vm_user=None, vm_pwd=None, filepath=None):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        creds = vim.vm.guest.NamePasswordAuthentication(username=vm_user, password=vm_pwd)
        self.si.content.guestOperationsManager.fileManager.DeleteFileInGuest(vm, creds, filepath)

    def readEnvVariableInVm(self, vm_id=None, vm_user=None, vm_pwd=None, names=None):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        creds = vim.vm.guest.NamePasswordAuthentication(username=vm_user, password=vm_pwd)
        results = self.si.content.guestOperationsManager.processManager.ReadEnvironmentVariableInGuest(vm, creds, names)

        variables = []
        for result in results:
            variables.append(result)
        return variables

    def executeProgramInVm(self, vm_id=None, vm_user=None, vm_pwd=None, program_path=None, arguments=None, timeout=180):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        pm = self.si.content.guestOperationsManager.processManager
        ps = vim.vm.guest.ProcessManager.ProgramSpec(programPath=program_path, arguments=arguments)
        creds = vim.vm.guest.NamePasswordAuthentication(username=vm_user, password=vm_pwd)
        res = pm.StartProgramInGuest(vm, creds, ps)
        execute_interval = 0
        pid_exitcode = -1

        if res > 0:
            pid_exitcode = pm.ListProcessesInGuest(vm, creds, [res]).pop().exitCode
            while (re.match('[^0-9]+', str(pid_exitcode))):
                if (execute_interval >= timeout):
                    break
                else:
                    time.sleep(5)
                    execute_interval += 5
                pid_exitcode = pm.ListProcessesInGuest(vm, creds, [res]).pop().exitCode
                if (re.match('[0-9]+', str(pid_exitcode))):
                    break

        out_json = {}
        out_json['exit_code'] = pid_exitcode
        return out_json

    def createDirectoryInVm(self, vm_id=None, vm_user=None, vm_pwd=None, dir_path=None):
        vm = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        creds = vim.vm.guest.NamePasswordAuthentication(username=vm_user, password=vm_pwd)
        try:
            self.si.content.guestOperationsManager.fileManager.MakeDirectoryInGuest(vm, creds, dir_path, True)
        except vim.fault.FileAlreadyExists:
            pass

    def downloadFile(self, vm_id=None, dest_path=None, file_name=None):
        vm_obj = self._get_obj_by_moid(vm_id, [vim.VirtualMachine])
        dc_obj = self._get_datacenter_obj(vm_obj)
        dc_path = self._get_datacenter_path(dc_obj)
        ds_name = file_name.split(']')[0][1:]
        vm_path = file_name.split(']')[1][1:]

        url = 'https://%s:%s/folder/%s?dcPath=%s&dsName=%s' % (self.auth_info['hostname'], self.auth_info['port'], vm_path, dc_path, ds_name)

        for retryCount in range(10):
            try:
                resp = requests.get(self._url_fix(url), verify=False, auth=(self.auth_info['username'], self.auth_info['password']))
                break
            except (requests.ConnectionError, requests.Timeout, requests.ConnectTimeout, requests.ReadTimeout):
                print('get connection error! retryCount = %d' % retryCount)
                time.sleep(2**retryCount)

        if isSuccessStatusCode(resp.status_code) == False:
            raise vim.fault.HostConnectFault(msg='Error while downloading file, status code: %d' % resp.status_code)

        dest_path = os.path.join(dest_path, os.path.basename(vm_path))
        with open(dest_path.encode('utf-8'), 'wb') as file:
            file.write(resp.content)

    def getMigrateCapability(self, host_id=None, host_obj=None):
        if host_obj == None:
            host_obj = self._get_obj_by_moid(host_id, [vim.HostSystem])
        return {
            'vmotion': host_obj.capability.vmotionSupported,
            'vcenter': self.api_type == 'VirtualCenter'
        }

    def getHostInterface(self, host_id=None):
        host_obj = self._get_obj_by_moid(host_id, [vim.HostSystem])
        return [vnic.spec.ip.ipAddress for vnic in host_obj.configManager.vmotionSystem.netConfig.candidateVnic]

    def runShell(self, command, timeout):
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(self.auth_info['hostname'],
            username=self.auth_info['username'],
            password=self.auth_info['password'])

        stdin, stdout, stderr = ssh.exec_command(command)

        if timeout > 0:
            elapsed = 0
            while not stdout.channel.exit_status_ready():
                if elapsed >= timeout:
                   raise vim.fault.Timedout(msg='execute [%s] timeout' % command)
                time.sleep(1)
                elapsed += 1

        status = stdout.channel.recv_exit_status()
        stdout = stdout.read().strip('\n')
        stderr = stderr.read().strip('\n')
        ssh.close()

        if status != 0:
            m = re.search(r'\((.*?)\) \{.*?msg = "(.*?)".*?\}', stderr, re.DOTALL)
            if m != None and m.lastindex == 2:
                raise eval(m.group(1))(msg=m.group(2))
            raise vim.fault.VimFault(msg=stderr)
        return stdout

    def listSession(self):
        print('session list of %s' % self.auth_info['hostname'])
        sessionList = self.si.content.sessionManager.sessionList
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.connect(('8.8.8.8', 80))
        hostAddress = sock.getsockname()[0]
        sock.close()

        for session in sessionList:
            if session.ipAddress == hostAddress:
                print(session.key, session.loginTime, session.userAgent)

    ####  util functions ###

    def _get_root_obj(self):
        return self.si.content.rootFolder

    def _get_obj_by_moid(self, moId, vimType=None, startObj=None):
        objs = self._get_objs(vimType, startObj)
        for obj in objs:
            if obj._moId == moId:
                return obj
        raise vim.fault.NotFound(msg='Cannot find [%s]' % moId)

    def _get_objs(self, vimType=None, startObj=None):
        if startObj is None:
            startObj = self._get_root_obj()

        container = self.si.content.viewManager.CreateContainerView(startObj, vimType, True)
        objs = [obj for obj in container.view]
        container.Destroy()
        return objs

    def _get_snapshot_obj(self, snapshots, moId):
        snapshot_obj_data = self._list_snapshot_obj_recursively(snapshots)
        for obj in snapshot_obj_data:
            if obj.snapshot._moId == moId:
                return obj
        raise vim.fault.NotFound(msg='Cannot find [%s]' % moId)

    def _list_snapshot_obj_recursively(self, snapshots):
        snapshot_obj_data = []
        for snapshot in snapshots:
            snapshot_obj_data.append(snapshot)
            snapshot_obj_data += self._list_snapshot_obj_recursively(snapshot.childSnapshotList)
        return snapshot_obj_data

    def _list_snapshot_recursively(self, snapshots):
        snapshot_data = []
        for snapshot in snapshots:
            snapshot_data.append(self._parse_snapshot_data(snapshot))
            snapshot_data += self._list_snapshot_recursively(snapshot.childSnapshotList)
        return snapshot_data

    def _parse_snapshot_data(self, snapshot):
        obj = {}
        obj['snap_id'] = snapshot.snapshot._moId
        obj['name'] = snapshot.name
        obj['desc'] = snapshot.description
        obj['create_time'] = int(time.mktime(snapshot.createTime.timetuple()))
        obj['virtual_disk'] = self._get_virtual_disk(snapshot.snapshot.config.hardware.device)
        return obj

    def _parse_question(self, question):
        choice = question.choice
        defaultIndex = (choice.defaultIndex if choice.defaultIndex else 0)
        defaultAnswer = choice.choiceInfo[defaultIndex]

        if defaultAnswer.label == 'button.uuid.copiedTheVM':
            for i in range(len(choice.choiceInfo)):
                if choice.choiceInfo[i].label == 'button.uuid.movedTheVM':
                    defaultAnswer = choice.choiceInfo[i]
                    break

        return {'question_id': question.id, 'default_answer': defaultAnswer.key}

    def _parse_datastore_data(self, datastore):
        info = datastore.info
        summary = datastore.summary

        out = {}
        out['name'] = summary.name
        out['accessible'] = summary.accessible
        out['capacity'] = summary.capacity
        out['freeSpace'] = summary.freeSpace
        out['datastore_id'] = datastore._moId

        if hasattr(info, 'nas'):
            out['volume_type'] = info.nas.type
            out['remote_host'] = info.nas.remoteHost
            out['remote_path'] = info.nas.remotePath
        if hasattr(info, 'vmfs'):
            out['volume_type'] = info.vmfs.type
        return out

    def _parse_host_data(self, host):
        out = {}
        out['host_id'] = host._moId
        out['name'] = host.name
        out['api_version'] = host.summary.config.product.apiVersion
        return out

    def _parse_vm_data(self, vm, detail=False, get_disk=False, get_view=False, view_cache={}):
        out = {}
        summary = vm.summary
        guest = summary.guest
        config = summary.config
        runtime = summary.runtime
        storage = summary.storage

        out['vm_id'] = vm._moId
        out['name'] = config.name
        out['os_name'] = config.guestFullName
        out['is_template'] = config.template
        out['instance_uuid'] = config.instanceUuid
        out['ram_mb'] = config.memorySizeMB
        out['num_cpu'] = config.numCpu
        out['num_vdisk'] = config.numVirtualDisks
        out['powered_on'] = runtime.powerState == vim.VirtualMachinePowerState.poweredOn
        out['vmtool_running'] = guest.toolsRunningStatus == vim.VirtualMachineToolsRunningStatus.guestToolsRunning
        out['connection_state'] = runtime.connectionState

        # for new VM list UI
        if get_view and vm.config:
            out['storage_size'] = storage.committed

            host = runtime.host
            if view_cache.get(host._moId) == None:
                view_cache[host._moId] = host.name if self.api_type == 'VirtualCenter' else ''
            out['hypervisor_name'] = view_cache[host._moId]

            folder = vm.parent if vm.parent != None else vm.resourcePool
            if view_cache.get(folder._moId) == None:
                view_cache[folder._moId] = self.getEntityPath(entity=folder, viewType='vim.VirtualMachine-Folder', attr='name')
            out['folder_path'] = view_cache[folder._moId]

        if get_disk and vm.config:
            out['virtual_disk'] = self._get_virtual_disk(vm.config.hardware.device)

        if detail and vm.config:
            out['files'] = []
            for file in vm.layoutEx.file:
                if file.name.endswith('vmx') or file.name.endswith('vmxf') or file.name.endswith('nvram'):
                    out['files'].append(file.name)

            dc = self._get_datacenter_obj(vm)
            if vm.parent:
                folder = vm.parent
                out['folder'] = {
                    'name': dc.name if folder.parent == dc else folder.name,
                    'folder_id': folder._moId,
                    'datacenter_id': dc._moId
                }
            else:
                out['folder'] = {
                    'name': dc.name,
                    'folder_id': dc.vmFolder._moId,
                    'datacenter_id': dc._moId
                }
            rp = vm.resourcePool
            out['resourcepool'] = {
                'name': rp.parent.name if isinstance(rp.parent, vim.ComputeResource) else rp.name,
                'is_vapp': (vm.parentVApp != None),
                'resourcepool_id': rp._moId
            }
            out['hypervisor'] = {
                'name': runtime.host.name,
                'host_id': runtime.host._moId,
                'standalone': type(runtime.host.parent) == vim.ComputeResource
            }
            out['datastore'] = []
            for datastore in vm.datastore:
                out['datastore'].append({
                    'name': datastore.name,
                    'datastore_id': datastore._moId
                })
            out['network'] = []
            for network in vm.network:
                out['network'].append({
                    'name': network.name,
                    'network_id': network._moId
                })
            out['hw_version'] = (vm.config and vm.config.version.split('-')[1])

        return out

    def _get_virtual_disk(self, devices):
        def _get_virtual_disk_filename_root(backing):
            while backing.parent:
                backing = backing.parent
            return backing.fileName

        disk_data = []
        device_objs = { device.key:device for device in devices }

        for device in device_objs.itervalues():
            if type(device) != vim.vm.device.VirtualDisk:
                continue

            backing = device.backing
            if backing == None:
                continue

            disk_obj = {}
            disk_obj['key'] = device.key
            disk_obj['name'] = device.deviceInfo.label
            disk_obj['unit_number'] = device.unitNumber
            disk_obj['capacity'] = device.capacityInKB * 1024
            disk_obj['change_id'] = backing.changeId
            disk_obj['uuid'] = backing.uuid
            disk_obj['thin_provision'] = backing.thinProvisioned
            disk_obj['eager_scrub'] = backing.eagerlyScrub
            disk_obj['filename'] = backing.fileName
            disk_obj['filename_root'] = _get_virtual_disk_filename_root(backing)
            controller_obj = device_objs[device.controllerKey]
            disk_obj['bus_number'] = controller_obj.busNumber
            disk_obj['controller_type'] = type(controller_obj).__name__
            disk_data.append(disk_obj)

        return disk_data

    def _get_datacenter_obj(self, obj):
        while type(obj) != vim.Datacenter:
            if hasattr(obj, 'parent') and obj.parent:
                obj = obj.parent
            if hasattr(obj, 'parentVApp') and obj.parentVApp:
                obj = obj.parentVApp
        return obj

    def _get_datacenter_path(self, obj):
        paths = []
        while obj != self._get_root_obj():
            paths.append(obj.name)
            obj = obj.parent
        return '/'.join(reversed(paths))

    def _url_fix(self, url, charset='utf-8'):
        if isinstance(url, unicode):
            url = url.encode(charset, 'ignore')
        scheme, netloc, path, qs, anchor = urlparse.urlsplit(url)
        path = urllib2.quote(path)
        qs = urllib2.quote(qs, ':&=')
        return urlparse.urlunsplit((scheme, netloc, path, qs, anchor))

    def _wait_for_task(self, task, timeout=2147483648):
        now = time.time()

        while True:
            state = task.info.state
            if (time.time() - now) > timeout:
                task.CancelTask()
                raise vim.fault.Timedout(msg='task timeout. task id [%s]' % task.info.key)
            if state == 'success':
                break
            elif state == 'error':
                raise task.info.error
            elif state in ('queued', 'running'):
                pass
            else:
                task.CancelTask()
                raise vim.fault.InvalidState(msg='unknown task state [%s]' % state)
            time.sleep(1)

        return task.info.result

    def _get_task(self, task_id):
        return self.task_pool[task_id]

    def _put_task(self, task):
        self.task_pool[task.info.key] = task
        return task.info.key

    def _del_task(self, task_id):
        del self.task_pool[task_id]

# run daemon

if __name__ == '__main__':
    serverPool = {}

    def GetServer(inventory):
        inventoryId = inventory['inventory_id']
        hostname    = inventory['host_name']
        username    = inventory['auth_user']
        password    = inventory['auth_password']
        thumbprint  = inventory['auth_thumbprint']
        port        = inventory['port_webapi']

        oldInventory, server = serverPool.get(inventoryId) or (None, None)

        if server == None or \
           (oldInventory['host_name'] != hostname or \
            oldInventory['auth_user'] != username or \
            oldInventory['auth_password'] != password or \
            oldInventory['auth_thumbprint'] != thumbprint or \
            oldInventory['port_webapi'] != port):

            server = vSphereServer(hostname, username, password, thumbprint, port)
            serverPool[inventoryId] = (inventory, server)

        return server

    def HandleRequest(request):
        command = request['command']
        argument = request['argument']
        inventory = request['inventory']
        server = GetServer(inventory)
        return server.dispatch(command, argument)

    def HandleConnection(connection, address):
        try:
            while True:
                data = connection.recv(4)
                if len(data) != 4: break

                n = sum([ord(data[i])*(256**i) for i in range(4)])

                data = connection.recv(n)
                if len(data) != n: break

                data = json.dumps(HandleRequest(json.loads(data)))
                n = len(data)
                data = ''.join([chr(n/(256**i)%256) for i in range(4)]) + data

                connection.sendall(data)
        finally:
            connection.close()

    def HandleSignal(sig, frame):
        if sig != signal.SIGPIPE:
            serverPool.clear()
            os._exit(0)

    print('vSphere Server Daemon Start! ^.^')

    try:
        os.unlink(VSPHERE_SERVER_SOCKET_PATH)
    except Exception:
        pass

    try:
        os.makedirs(os.path.dirname(VSPHERE_SERVER_SOCKET_PATH))
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    sock.bind(VSPHERE_SERVER_SOCKET_PATH)
    sock.listen(16)

    os.chown(VSPHERE_SERVER_SOCKET_PATH, int(os.environ['PKG_UID']), int(os.environ['PKG_GID']))
    os.chmod(VSPHERE_SERVER_SOCKET_PATH, stat.S_IRUSR + stat.S_IWUSR)

    for sig in (signal.SIGINT, signal.SIGTERM, signal.SIGQUIT, signal.SIGPIPE):
        signal.signal(sig, HandleSignal)

    while True:
        threading.Thread(target=HandleConnection, args=sock.accept()).start()
